
using OpenTelemetryAPI
using OpenTelemetrySDK
using OpenTelemetryExporterPrometheus 
using OpenTelemetryExporterOtlpProtoGrpc
using Random

struct AsyncCallbacks
    a_usage::Function
    b_usage::Function
    c_usage::Function
    a_running::Function
    b_running::Function
    a_sync_last::Function
end

mutable struct Observe
    me::PrometheusExporter
    m::Dict{String, AbstractInstrument}
    jr::Dict{String, Dict{String, UInt64}}
    cb::AsyncCallbacks
    tr::Tracer
    function Observe(trace_enabled, trace_scheme, trace_host, trace_port)
        metric_exporter = PrometheusExporter(; host="0.0.0.0", port=9967)  # port=9966
        if trace_enabled
            trace_exporter = OtlpProtoGrpcTraceExporter(;
                scheme=trace_scheme,
                host=trace_host,
                port=trace_port
            )
            global_tracer_provider!(
                TracerProvider(;
                    span_processor=CompositSpanProcessor(
                        SimpleSpanProcessor(
                            trace_exporter
                        ),
                    )
                ),
            )
            tracer = Tracer()
        else
            tracer = Tracer(; provider=OpenTelemetryAPI.DUMMY_TRACER_PROVIDER)
        end
        metric_provider = MeterProvider(;
            views=[
                View(
                    "a_seconds";
                    instrument_name="a_seconds",
                    aggregation=HistogramAgg{Float64}(; boundaries=Tuple(0.0:5.0:25.0)),
                ),
                View(
                    "b_seconds";
                    instrument_name="b_seconds",
                    aggregation=HistogramAgg{Float64}(; boundaries=Tuple(0.0:1.0:5.0)),
                ),
                View(
                    "c_seconds";
                    instrument_name="c_seconds",
                    aggregation=HistogramAgg{Float64}(; boundaries=Tuple(0.0:1.0:5.0)),
                ),
                View(
                    "d_seconds";
                    instrument_name="d_seconds",
                    aggregation=HistogramAgg{Float64}(; boundaries=Tuple(0.0:25.0:100.0)),
                ),
                View(
                    "e_seconds";
                    instrument_name="e_seconds",
                    aggregation=HistogramAgg{Float64}(; boundaries=Tuple(0.0:2.5:10.0)),
                ),
                View(
                    "f_seconds";
                    instrument_name="f_seconds",
                    aggregation=HistogramAgg{Float64}(; boundaries=Tuple(0.0:75.0:300.0)),
                ),
                View(
                    "g_seconds";
                    instrument_name="g_seconds",
                    aggregation=HistogramAgg{Float64}(; boundaries=Tuple(0.0:75.0:300.0)),
                ),
                View(
                    "h_seconds";
                    instrument_name="h_seconds",
                    aggregation=HistogramAgg{Float64}(; boundaries=Tuple(0.0:15.0:60.0)),
                ),
                View(
                    "i_seconds";
                    instrument_name="i_seconds",
                    aggregation=HistogramAgg{Float64}(; boundaries=Tuple(0.0:360.0:3600.0)),
                ),
                View(;
                    instrument_name="*"
                ),
            ],
        )
        MetricReader(metric_provider, metric_exporter)
        m = Meter("App"; provider=metric_provider)
        observe = new(
            metric_exporter,
            Dict(),
            Dict(),
            AsyncCallbacks(
                function ()
                    cpu = rand()
                    Measurement(cpu)
                end,
                function ()
                    gpu = rand()
                    Measurement(gpu)
                end,
                function ()
                    mem = rand()
                    Measurement(mem * 1_073_741_824.0)
                end,
                function ()
                    Measurement(floor(Int, rand() * 100))
                end,
                function ()
                    Measurement(floor(Int, rand() * 100))
                end,
                function ()
                    Measurement(floor(Int, rand() * 100))
                end,
            ),
            tracer,
        )
        observe.m = Dict(
            "a_seconds" => Histogram{Float64}("a_seconds", m),
            "b_seconds" => Histogram{Float64}("b_seconds", m),
            "c_seconds" => Histogram{Float64}("c_seconds", m),
            "d_seconds" => Histogram{Float64}("d_seconds", m),
            "e_seconds" => Histogram{Float64}("e_seconds", m),
            "f_seconds" => Histogram{Float64}("f_seconds", m),
            "g_seconds" => Histogram{Float64}("g_seconds", m),
            "h_seconds" => Histogram{Float64}("h_seconds", m),
            "i_seconds" => Histogram{Float64}("i_seconds", m),
            "a_usage" => ObservableGauge{Float64}(observe.cb.a_usage, "a_usage", m),
            "b_usage" => ObservableGauge{Float64}(observe.cb.b_usage, "b_usage", m),
            "c_usage" => ObservableGauge{Float64}(observe.cb.c_usage, "c_usage", m),
            "a_running" => ObservableGauge{UInt64}(observe.cb.a_running, "a_running", m),
            "b_running" => ObservableGauge{UInt64}(observe.cb.b_running, "b_running", m),
            "a_sync_last" => ObservableGauge{UInt64}(observe.cb.a_sync_last, "a_sync_last", m),
            "a_status" => Counter{UInt64}("a_status", m),
        )
        observe
    end
end

function addmeasure(o::Observe, name, dur, name2)
    o.m[name](dur; prod=name2)
end

function metricsspam(o::Observe)
    while true
        sleep(0.5)
        for m in [
            "a_seconds",
            "b_seconds",
            "c_seconds",
            "d_seconds",
            "e_seconds",
            "f_seconds",
            "g_seconds",
            "h_seconds",
            "i_seconds",
        ]
            addmeasure(o, m, rand(), rand(["A", "B", "C"]))
        end
    end

end

function init()
    t = @async begin
        o = try
            Observe(
                false,
                "http",
                "host.docker.internal",
                "4318",
            )
        catch ex
            @error(
                "exception initializing",
                exception = (ex, catch_backtrace())
            )
        end
        @async begin
            try
                metricsspam(o)
            catch ex
                @error(
                    "exception initializing",
                    exception = (ex, catch_backtrace())
                )
            end
        end

    end
    wait()
end

